//  Marco Di Mario (mkdm), 2016
// Karsten Marquas (kmrqus), 2017

function selectBylist(namesList) {
    var selectedObjects = moi.geometryDatabase.createObjectList();
    moi.geometryDatabase.deselectAll();
    for (var i = 0; i < namesList.length; i++)
    {
        moi.geometryDatabase.selectNamed(namesList[i]);
    }
    selectedObjects = moi.geometryDatabase.getSelectedObjects();

    moi.geometryDatabase.deselectAll();
    return selectedObjects;
}

function createFrameByLines(basePt, DirLineU, DirLineV) //mathpoint,object,object
{
    var uPtStart = DirLineU.getStartPt(); //Anfangspunkt der Kurve erzeugen
    var uPtEnd = DirLineU.getEndPt();
    var vPtStart = DirLineV.getStartPt();
    var vPtEnd = DirLineV.getEndPt();
    // create point object to array

    var dirU = moi.VectorMath.createPoint(
            uPtEnd.x - uPtStart.x,
            uPtEnd.y - uPtStart.y,
            uPtEnd.z - uPtStart.z
            );
    var dirV = moi.VectorMath.createPoint(
            vPtEnd.x - vPtStart.x,
            vPtEnd.y - vPtStart.y,
            vPtEnd.z - vPtStart.z
            );

    var outFrame = moi.VectorMath.createFrame(basePt, dirU, dirV);

    return outFrame;
}

function deleteObjects(nameArray)
{
    moi.geometryDatabase.deselectAll();
    if (nameArray.length > 0)
    {
        var deleteList = moi.geometryDatabase.createObjectList();
        for (var i = 0; i < nameArray.length; i++)
        {
            moi.geometryDatabase.selectNamed(nameArray[i]);
        }
        deleteList = moi.geometryDatabase.getSelectedObjects();

        var deleteOldFaceFac = moi.command.createFactory('delete');
        deleteOldFaceFac.setInput(0, deleteList);
        deleteOldFaceFac.calculate();
        deleteOldFaceFac.commit();
        ///// cleaning up faceList
        while (nameArray.length > 0)
        {
            nameArray.pop();
        }
    }
}

function NormalilzeVA(VA, val)
{
    var nav = new Array();
    var max = VA[0];
    for (var i = 0; i < VA.length; i++) {
        if (VA[i] > max)
            max = VA[i];
    }
    for (var i = 0; i < VA.length; i++) {

        if (max != 0)
            nav[i] = VA[i] / max * val;
    }
    return nav;
}
// kmrqus
function checkSMI(val)
{	//test for strictly monotonic increasing
    for (var i = 0; i < val.length - 2; i++) {
        if (val[i] > val[i + 1])
        {
            return false;
        }

        return true;
    }
}

// stolen from Max:-)
function crvlength(crvs)
{
    var len = 0.0;
    for (var i = 0; i < crvs.length; ++i)
        len += crvs.item(i).getLength();
    return len;
}
// example from http://stackoverflow.com/questions/6248666/how-to-generate-short-uid-like-ax4j9z-in-js
/**
 * function generates a UID string (unique identity) as a chain of random characters and numbers
 * 
 * @param {type} len
 * @returns {getUID.uidsout}
 */
function getUID(len) {
    var chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789',
            out = '';
    for (var i = 0, clen = chars.length; i < len; i++)
    {
        out += chars.substr(0 | Math.random() * clen, 1);
    }

// ensure that the uid is unique for this page
    return getUID.uids[out] ? getUID(len) : (getUID.uids[out] = out);
}
getUID.uids = {};

(function () {
/////////////////////
    function pathArray()
    {
        this.addInput("Rail", "objectlist"); //  [1] railcurve
        //a single value with the number of sections or 
        //an array with monotonic raising values between 0 and 1
        this.addInput("Div Rail", "numarray"); //  [2] Division Number of Rail a single value with the number of sections
        //or an array with monotonic raising values between 0 and 1
        this.addOutput("Out", "pointarray"); //[0]

        this.properties = {rotationMode: ["Freeform", "Freeform", "Flat"], DivNumRail: [6]};

        this.Angle = new Array();
    }
    pathArray.title = "Path_Array";
    pathArray.desc = "Path_Array";
    pathArray.prototype.onAdded = function ()
    {
    }

    pathArray.prototype.onPropertyChange = function ()
    {
    }
    pathArray.prototype.onExecute = function ()
    {
        this.profilePts = moi.geometryDatabase.createObjectList();
        var rail = this.getInputData(0, moi.geometryDatabase.createObjectList());
        this.properties.DivNumRail = this.getInputData(1, this.properties.DivNumRail);

        if (rail.length > 0)
        {
////////////////////////////// Create Base Frame //////////////////////////////
            var startPt = rail.item(0).getStartPt();
            var startPtObject = factory('point', startPt);
            /// creating a pointarray on the curve
            var ptArray = factory('arraycurve', startPtObject, rail.item(0), 'NumItems', 500);
            /// calculating the components of 3 secants (approx. tangents) under the second (index1), third(index2) and fourth point (index3)
            var dx_1 = ptArray.item(1).pt.x - startPtObject.item(0).pt.x;
            var dy_1 = ptArray.item(1).pt.y - startPtObject.item(0).pt.y;
            var dz_1 = ptArray.item(1).pt.z - startPtObject.item(0).pt.z;
            var dx_2 = ptArray.item(2).pt.x - ptArray.item(0).pt.x;
            var dy_2 = ptArray.item(2).pt.y - ptArray.item(0).pt.y;
            var dz_2 = ptArray.item(2).pt.z - ptArray.item(0).pt.z;
            var dx_3 = ptArray.item(3).pt.x - ptArray.item(1).pt.x;
            var dy_3 = ptArray.item(3).pt.y - ptArray.item(1).pt.y;
            var dz_3 = ptArray.item(3).pt.z - ptArray.item(1).pt.z;

//Lagrange interpolation with fixed values the Indices are the "x"-values (t-values) here - 0 is the variable of the startpoint
            var dx_0 = dx_1 * (0 - 2) * (0 - 3) / ((1 - 2) * (1 - 3)) + dx_2 * (0 - 1) * (0 - 3) / ((2 - 1) * (2 - 3)) + dx_3 * (0 - 1) * (0 - 2) / ((3 - 1) * (3 - 2));
            var dy_0 = dy_1 * (0 - 2) * (0 - 3) / ((1 - 2) * (1 - 3)) + dy_2 * (0 - 1) * (0 - 3) / ((2 - 1) * (2 - 3)) + dy_3 * (0 - 1) * (0 - 2) / ((3 - 1) * (3 - 2));
            var dz_0 = dz_1 * (0 - 2) * (0 - 3) / ((1 - 2) * (1 - 3)) + dz_2 * (0 - 1) * (0 - 3) / ((2 - 1) * (2 - 3)) + dz_3 * (0 - 1) * (0 - 2) / ((3 - 1) * (3 - 2));
            var frameOrigin = moi.VectorMath.createPoint(
                    startPtObject.item(0).pt.x,
                    startPtObject.item(0).pt.y,
                    startPtObject.item(0).pt.z
                    );

            var n_vec = moi.VectorMath.createPoint(
                    dx_0,
                    dy_0,
                    dz_0
                    );
            var n0_vec = scaleVec(n_vec, 1 / magnitude(n_vec));

            var dHNF = dotProd(n0_vec, frameOrigin); //ScalarProd from startpoint and tangent( normal vector to get distance of Hesse normalform 
            var abs_dx_0 = Math.abs(dx_0);
            var abs_dy_0 = Math.abs(dy_0);
            var abs_dz_0 = Math.abs(dz_0);
            if (abs_dx_0 >= abs_dy_0 && abs_dx_0 >= abs_dz_0)
            {
                var u_help = moi.VectorMath.createPoint(
                        frameOrigin.x,
                        frameOrigin.y + 1,
                        frameOrigin.z
                        );
                /*                var v_help = moi.VectorMath.createPoint(
                 frameOrigin.x,
                 frameOrigin.y,
                 frameOrigin.z + 1
                 );*/
            }
            if (abs_dy_0 > abs_dx_0 && abs_dy_0 >= abs_dz_0)
            {
                var u_help = moi.VectorMath.createPoint(
                        frameOrigin.x + 1,
                        frameOrigin.y,
                        frameOrigin.z
                        );
            }
            if (abs_dz_0 > abs_dx_0 && abs_dz_0 > abs_dy_0)
            {
                var u_help = moi.VectorMath.createPoint(
                        frameOrigin.x + 1,
                        frameOrigin.y,
                        frameOrigin.z
                        );
            }

            var uLambda = (-(n0_vec.x * u_help.x + n0_vec.y * u_help.y + n0_vec.z * u_help.z) + dHNF) / (n0_vec.x * n0_vec.x + n0_vec.y * n0_vec.y + n0_vec.z * n0_vec.z);
//            var vLambda = (-(n0_vec.x * v_help.x + n0_vec.y * v_help.y + n0_vec.z * v_help.z) + dHNF) / (n0_vec.x * n0_vec.x + n0_vec.y * n0_vec.y + n0_vec.z * n0_vec.z);

            var dirU = moi.VectorMath.createPoint(
                    u_help.x + uLambda * n0_vec.x - frameOrigin.x,
                    u_help.y + uLambda * n0_vec.y - frameOrigin.y,
                    u_help.z + uLambda * n0_vec.z - frameOrigin.z
                    );

            var dirV = crossProd(n0_vec, dirU);

            var bFrame = moi.VectorMath.createFrame(frameOrigin, dirU, dirV);
////////////////////////////////End Create Base Frame //////////////////////////////
////////////get points back from the frame to array them/////////////////////////////////
            var bframeOrigin = factory('point', bFrame.origin);
            var x_axis = moi.VectorMath.createPoint(bFrame.origin.x + bFrame.xaxis.x, bFrame.origin.y + bFrame.xaxis.y, bFrame.origin.z + bFrame.xaxis.z);
            var bframeXAxis = factory('point', x_axis);
            var y_axis = moi.VectorMath.createPoint(bFrame.origin.x + bFrame.yaxis.x, bFrame.origin.y + bFrame.yaxis.y, bFrame.origin.z + bFrame.yaxis.z);
            var bframeYAxis = factory('point', y_axis);
///////////end Get points back from the frame to array them/////////////////////////////////
/////////workaround converted Points used for generate an array of frames
            var bFramePts=moi.geometryDatabase.createObjectList();
            bFramePts.addObject(bframeOrigin.item(0));
            bFramePts.addObject(bframeXAxis.item(0));
            bFramePts.addObject(bframeYAxis.item(0));
            if (this.properties.DivNumRail.length > 1)
            {
                var framePts = CurveDiv(rail.item(0), this.properties.DivNumRail, bFramePts, this.properties.rotationMode[0]);                
            } else {
                var framePts = factory('arraycurve', bFramePts, rail.item(0), 'NumItems', this.properties.DivNumRail[0], null, this.properties.rotationMode[0]);
            }
/////////end workaround converted Points used for generate an array of frames
            var outframe = new pointArray(false);
            outframe.pushFrame(bFrame);
            for (var i = 2; i < framePts.length; i=i+3)
            {
                var xdir = moi.VectorMath.createPoint(
                        framePts.item(i-1).pt.x - framePts.item(i-2).pt.x,
                        framePts.item(i-1).pt.y - framePts.item(i-2).pt.y,
                        framePts.item(i-1).pt.z - framePts.item(i-2).pt.z);
                var ydir = moi.VectorMath.createPoint(
                        framePts.item(i).pt.x - framePts.item(i-2).pt.x,
                        framePts.item(i).pt.y - framePts.item(i-2).pt.y,
                        framePts.item(i).pt.z - framePts.item(i-2).pt.z);
                var temp = moi.VectorMath.createFrame(framePts.item(i-2).pt, xdir, ydir);
                outframe.pushFrame(temp);
            }
            this.setOutputData(0, outframe);
        }
    }
    pathArray.prototype.onDrawBackground = function (ctx) {
        this.outputs[0].label = this.test;
    }

    LiteGraph.registerNodeType("Points2/PathPoints", pathArray);
///
//
    function mpathArray()
    {
        this.addInput("Rail", "objectlist"); //  [1] railcurve
        //a single value with the number of sections or 
        //an array with monotonic raising values between 0 and 1
        this.addInput("Div Rail", "numarray"); //  [2] Division Number of Rail a single value with the number of sections
        //or an array with monotonic raising values between 0 and 1
        this.addOutput("Out", "pointarray"); //[0]

        this.properties = {ptOrder: ["along", "along", "cross"],rotationMode: ["Freeform", "Freeform", "Flat"], DivNumRail: [6]};
        this.Angle = new Array();
    }
    mpathArray.title = "mPath_Array";
    mpathArray.desc = "mPath_Array";
    mpathArray.prototype.onAdded = function ()
    {
    }

    mpathArray.prototype.onPropertyChange = function ()
    {
    }
    mpathArray.prototype.onExecute = function ()
    {
        this.profilePts = moi.geometryDatabase.createObjectList();
        var rails = this.getInputData(0, moi.geometryDatabase.createObjectList());
        this.properties.DivNumRail = this.getInputData(1, this.properties.DivNumRail);
        var frames = new pointArray(false);
//        moi.ui.alert(rails.length,"rails.length");
//moi.ui.alert(rails.numCurves,"rails.numCurves");
        for (var i=0;i<rails.numCurves;i++)
        {   //moi.ui.alert(i,"loop");
////////////////////////////// Create Base Frame //////////////////////////////
            var startPt = rails.item(i).getStartPt();
            var startPtObject = factory('point', startPt);
            /// creating a pointarray on the curve
            var ptArray = factory('arraycurve', startPtObject, rails.item(i), 'NumItems', 500);
            /// calculating the components of 3 secants (approx. tangents) under the second (index1), third(index2) and fourth point (index3)
            var dx_1 = ptArray.item(1).pt.x - startPtObject.item(0).pt.x;
            var dy_1 = ptArray.item(1).pt.y - startPtObject.item(0).pt.y;
            var dz_1 = ptArray.item(1).pt.z - startPtObject.item(0).pt.z;
            var dx_2 = ptArray.item(2).pt.x - ptArray.item(0).pt.x;
            var dy_2 = ptArray.item(2).pt.y - ptArray.item(0).pt.y;
            var dz_2 = ptArray.item(2).pt.z - ptArray.item(0).pt.z;
            var dx_3 = ptArray.item(3).pt.x - ptArray.item(1).pt.x;
            var dy_3 = ptArray.item(3).pt.y - ptArray.item(1).pt.y;
            var dz_3 = ptArray.item(3).pt.z - ptArray.item(1).pt.z;

//Lagrange interpolation with fixed values the Indices are the "x"-values (t-values) here - 0 is the variable of the startpoint
            var dx_0 = dx_1 * (0 - 2) * (0 - 3) / ((1 - 2) * (1 - 3)) + dx_2 * (0 - 1) * (0 - 3) / ((2 - 1) * (2 - 3)) + dx_3 * (0 - 1) * (0 - 2) / ((3 - 1) * (3 - 2));
            var dy_0 = dy_1 * (0 - 2) * (0 - 3) / ((1 - 2) * (1 - 3)) + dy_2 * (0 - 1) * (0 - 3) / ((2 - 1) * (2 - 3)) + dy_3 * (0 - 1) * (0 - 2) / ((3 - 1) * (3 - 2));
            var dz_0 = dz_1 * (0 - 2) * (0 - 3) / ((1 - 2) * (1 - 3)) + dz_2 * (0 - 1) * (0 - 3) / ((2 - 1) * (2 - 3)) + dz_3 * (0 - 1) * (0 - 2) / ((3 - 1) * (3 - 2));
            var frameOrigin = moi.VectorMath.createPoint(
                    startPtObject.item(0).pt.x,
                    startPtObject.item(0).pt.y,
                    startPtObject.item(0).pt.z
                    );

            var n_vec = moi.VectorMath.createPoint(
                    dx_0,
                    dy_0,
                    dz_0
                    );
            var n0_vec = scaleVec(n_vec, 1 / magnitude(n_vec));

            var dHNF = dotProd(n0_vec, frameOrigin); //ScalarProd from startpoint and tangent( normal vector to get distance of Hesse normalform 
            var abs_dx_0 = Math.abs(dx_0);
            var abs_dy_0 = Math.abs(dy_0);
            var abs_dz_0 = Math.abs(dz_0);
            if (abs_dx_0 >= abs_dy_0 && abs_dx_0 >= abs_dz_0)
            {
                var u_help = moi.VectorMath.createPoint(
                        frameOrigin.x,
                        frameOrigin.y + 1,
                        frameOrigin.z
                        );
                /*                var v_help = moi.VectorMath.createPoint(
                 frameOrigin.x,
                 frameOrigin.y,
                 frameOrigin.z + 1
                 );*/
            }
            if (abs_dy_0 > abs_dx_0 && abs_dy_0 >= abs_dz_0)
            {
                var u_help = moi.VectorMath.createPoint(
                        frameOrigin.x + 1,
                        frameOrigin.y,
                        frameOrigin.z
                        );
            }
            if (abs_dz_0 > abs_dx_0 && abs_dz_0 > abs_dy_0)
            {
                var u_help = moi.VectorMath.createPoint(
                        frameOrigin.x + 1,
                        frameOrigin.y,
                        frameOrigin.z
                        );
            }

            var uLambda = (-(n0_vec.x * u_help.x + n0_vec.y * u_help.y + n0_vec.z * u_help.z) + dHNF) / (n0_vec.x * n0_vec.x + n0_vec.y * n0_vec.y + n0_vec.z * n0_vec.z);
//            var vLambda = (-(n0_vec.x * v_help.x + n0_vec.y * v_help.y + n0_vec.z * v_help.z) + dHNF) / (n0_vec.x * n0_vec.x + n0_vec.y * n0_vec.y + n0_vec.z * n0_vec.z);

            var dirU = moi.VectorMath.createPoint(
                    u_help.x + uLambda * n0_vec.x - frameOrigin.x,
                    u_help.y + uLambda * n0_vec.y - frameOrigin.y,
                    u_help.z + uLambda * n0_vec.z - frameOrigin.z
                    );
            /*            var dirV = moi.VectorMath.createPoint(
             v_help.x + vLambda * n0_vec.x - frameOrigin.x,
             v_help.y + vLambda * n0_vec.y - frameOrigin.y,
             v_help.z + vLambda * n0_vec.z - frameOrigin.z
             );*/
            var dirV = crossProd(n0_vec, dirU);

            var bFrame = moi.VectorMath.createFrame(frameOrigin, dirU, dirV);
////////////////////////////////End Create Base Frame //////////////////////////////
////////////get points back from the frame to array them/////////////////////////////////
            var bframeOrigin = factory('point', bFrame.origin);
            var x_axis = moi.VectorMath.createPoint(bFrame.origin.x + bFrame.xaxis.x, bFrame.origin.y + bFrame.xaxis.y, bFrame.origin.z + bFrame.xaxis.z);
            var bframeXAxis = factory('point', x_axis);
            var y_axis = moi.VectorMath.createPoint(bFrame.origin.x + bFrame.yaxis.x, bFrame.origin.y + bFrame.yaxis.y, bFrame.origin.z + bFrame.yaxis.z);
            var bframeYAxis = factory('point', y_axis);
///////////end Get points back from the frame to array them/////////////////////////////////
/////////workaround converted Points used for generate an array of frames
            var bFramePts=moi.geometryDatabase.createObjectList();
            bFramePts.addObject(bframeOrigin.item(0));
            bFramePts.addObject(bframeXAxis.item(0));
            bFramePts.addObject(bframeYAxis.item(0));
            if (this.properties.DivNumRail.length > 1)
            {
                var framePts = CurveDiv(rails.item(i), this.properties.DivNumRail, bFramePts, this.properties.rotationMode[0]);                
            } else {
                var framePts = factory('arraycurve', bFramePts, rails.item(i), 'NumItems', this.properties.DivNumRail[0], null, this.properties.rotationMode[0]);
            }
/////////end workaround converted Points used for generate an array of frames
  //          var outframe = new pointArray(false);
            frames.pushFrame(bFrame);
            for (var u = 2; u < framePts.length; u=u+3)
            {
                var xdir = moi.VectorMath.createPoint(
                        framePts.item(u-1).pt.x - framePts.item(u-2).pt.x,
                        framePts.item(u-1).pt.y - framePts.item(u-2).pt.y,
                        framePts.item(u-1).pt.z - framePts.item(u-2).pt.z);
                var ydir = moi.VectorMath.createPoint(
                        framePts.item(u).pt.x - framePts.item(u-2).pt.x,
                        framePts.item(u).pt.y - framePts.item(u-2).pt.y,
                        framePts.item(u).pt.z - framePts.item(u-2).pt.z);
                var frm = moi.VectorMath.createFrame(framePts.item(u-2).pt, xdir, ydir);
                frames.pushFrame(frm);
//          moi.ui.alert(frames.getLength(),"frames.getLength()");
            }

//            this.setOutputData(0, outframe);
        } //end of curves routine
        if(this.properties.ptOrder[0]==="along"){
            frames.xlength=(this.properties.DivNumRail.length>1)? this.properties.DivNumRail.length:this.properties.DivNumRail[0];
            frames.ylength=1;
            frames.zlength=rails.length;
//                            moi.ui.alert(rails.length,"rails.length");
//                            moi.ui.alert(frames.getLength(),"frames.getLength()");
            this.setOutputData(0, frames);
                }else{        
                    var crossframes=new pointArray(false);
                    var z=(this.properties.DivNumRail.length>1)? this.properties.DivNumRail.length:this.properties.DivNumRail[0];
                    for(var i=0;i<z;i++){
                        for(var j=0;j<=rails.length;j++){
                        crossframes.concat(frames.getElement(j*z+i));
                        }
                    crossframes.xlength=rails.length;
                    crossframes.ylength=1;
                    crossframes.zlength=(this.properties.DivNumRail.length>1)? this.properties.DivNumRail.length:this.properties.DivNumRail[0];
                    }

                    this.setOutputData(0, crossframes);
                }
    }
    mpathArray.prototype.onDrawBackground = function (ctx) {
        this.outputs[0].label = this.test;
    }

    LiteGraph.registerNodeType("Points2/mPathPoints", mpathArray);

//InterpCurve
    function getCurveStartFrame()
    {
        this.addInput("In", "objectlist");
        this.addOutput("Out", "objectlist");
        this.addOutput("Start Frame", "pointarray");
    }
    getCurveStartFrame.title = "getCrvSFrame";
    getCurveStartFrame.desc = "getCurveStartFrame";

    getCurveStartFrame.prototype.onExecute = function ()
    {
        var output = this.getInputData(0, moi.geometryDatabase.createObjectList());
        var outframe = new pointArray(false);
        this.setOutputData(0, output);
        var fcrvList = factory('flip', output);
//////////////calculating the frame /////////////
        ///evaluating the startpoint of the curve
        for(var i=0;i<output.length;i++)
        {               // create point object to array
            var startPtObject = factory('point', output.item(i).getStartPt());
            var endPt = factory('point', fcrvList.item(i).getStartPt());
            /// creating a pointarray on the curve
            var fcrvObj = fcrvList.item(i);
            var hlpPt1 = factory('arraycurve', endPt, fcrvObj, 'Distance', null, fcrvObj.getLength() *0.999, 'Freeform');
            var hlpPt2 = factory('arraycurve', endPt, fcrvObj, 'Distance', null, fcrvObj.getLength() *0.998, 'Freeform');
            var hlpPt3 = factory('arraycurve', endPt, fcrvObj, 'Distance', null, fcrvObj.getLength() *0.997, 'Freeform');
            var hlpPt4 = factory('arraycurve', endPt, fcrvObj, 'Distance', null, fcrvObj.getLength() *0.996, 'Freeform');
            var ptArray=moi.geometryDatabase.createObjectList();
            ptArray.addObject(hlpPt1.item(0));
            ptArray.addObject(hlpPt2.item(0));
            ptArray.addObject(hlpPt3.item(0));
            ptArray.addObject(hlpPt4.item(0));

            /// calculating the components of 3 secants (approx. tangents) under the second (index1), third(index2) and fourth point (index3)
            var dx_1 = ptArray.item(1).pt.x - startPtObject.item(0).pt.x;
            var dy_1 = ptArray.item(1).pt.y - startPtObject.item(0).pt.y;
            var dz_1 = ptArray.item(1).pt.z - startPtObject.item(0).pt.z;
            var dx_2 = ptArray.item(2).pt.x - ptArray.item(0).pt.x;
            var dy_2 = ptArray.item(2).pt.y - ptArray.item(0).pt.y;
            var dz_2 = ptArray.item(2).pt.z - ptArray.item(0).pt.z;
            var dx_3 = ptArray.item(3).pt.x - ptArray.item(1).pt.x;
            var dy_3 = ptArray.item(3).pt.y - ptArray.item(1).pt.y;
            var dz_3 = ptArray.item(3).pt.z - ptArray.item(1).pt.z;

//Lagrange interpolation with fixed values the Indices are the "x"-values (t-values) here - 0 is the variable of the startpoint
            var dx_0 = dx_1 * (0 - 2) * (0 - 3) / ((1 - 2) * (1 - 3)) + dx_2 * (0 - 1) * (0 - 3) / ((2 - 1) * (2 - 3)) + dx_3 * (0 - 1) * (0 - 2) / ((3 - 1) * (3 - 2));
            var dy_0 = dy_1 * (0 - 2) * (0 - 3) / ((1 - 2) * (1 - 3)) + dy_2 * (0 - 1) * (0 - 3) / ((2 - 1) * (2 - 3)) + dy_3 * (0 - 1) * (0 - 2) / ((3 - 1) * (3 - 2));
            var dz_0 = dz_1 * (0 - 2) * (0 - 3) / ((1 - 2) * (1 - 3)) + dz_2 * (0 - 1) * (0 - 3) / ((2 - 1) * (2 - 3)) + dz_3 * (0 - 1) * (0 - 2) / ((3 - 1) * (3 - 2));
            var frameOrigin = moi.VectorMath.createPoint(
                    startPtObject.item(0).pt.x,
                    startPtObject.item(0).pt.y,
                    startPtObject.item(0).pt.z
                    );

            var n_vec = moi.VectorMath.createPoint(
                    dx_0,
                    dy_0,
                    dz_0
                    );
            var n0_vec = scaleVec(n_vec, 1 / magnitude(n_vec));

            var dHNF = dotProd(n0_vec, frameOrigin); //ScalarProd from startpoint and tangent( normal vector to get distance of Hesse normalform 

            var abs_dx_0 = Math.abs(dx_0);
            var abs_dy_0 = Math.abs(dy_0);
            var abs_dz_0 = Math.abs(dz_0);
            if (abs_dx_0 >= abs_dy_0 && abs_dx_0 >= abs_dz_0)
            {
                var u_help = moi.VectorMath.createPoint(
                        frameOrigin.x,
                        frameOrigin.y + 1,
                        frameOrigin.z
                        );
                var v_help = moi.VectorMath.createPoint(
                        frameOrigin.x,
                        frameOrigin.y,
                        frameOrigin.z + 1
                        );
            }
            if (abs_dy_0 > abs_dx_0 && abs_dy_0 >= abs_dz_0)
            {
                var u_help = moi.VectorMath.createPoint(
                        frameOrigin.x + 1,
                        frameOrigin.y,
                        frameOrigin.z
                        );
                var v_help = moi.VectorMath.createPoint(
                        frameOrigin.x,
                        frameOrigin.y,
                        frameOrigin.z + 1
                        );
            }
            if (abs_dz_0 > abs_dx_0 && abs_dz_0 > abs_dy_0)
            {
                var u_help = moi.VectorMath.createPoint(
                        frameOrigin.x + 1,
                        frameOrigin.y,
                        frameOrigin.z
                        );
                var v_help = moi.VectorMath.createPoint(
                        frameOrigin.x,
                        frameOrigin.y + 1,
                        frameOrigin.z
                        );
            }

            var uLambda = (-(n0_vec.x * u_help.x + n0_vec.y * u_help.y + n0_vec.z * u_help.z) + dHNF) / (n0_vec.x * n0_vec.x + n0_vec.y * n0_vec.y + n0_vec.z * n0_vec.z);
            var vLambda = (-(n0_vec.x * v_help.x + n0_vec.y * v_help.y + n0_vec.z * v_help.z) + dHNF) / (n0_vec.x * n0_vec.x + n0_vec.y * n0_vec.y + n0_vec.z * n0_vec.z);

            var dirU = moi.VectorMath.createPoint(
                    u_help.x + uLambda * n0_vec.x - frameOrigin.x,
                    u_help.y + uLambda * n0_vec.y - frameOrigin.y,
                    u_help.z + uLambda * n0_vec.z - frameOrigin.z
                    );
/*            var dirV = moi.VectorMath.createPoint(
                    v_help.x + vLambda * n0_vec.x - frameOrigin.x,
                    v_help.y + vLambda * n0_vec.y - frameOrigin.y,
                    v_help.z + vLambda * n0_vec.z - frameOrigin.z
                    );*/
            var dirV=crossProd(n0_vec,dirU);

            var startFrame = moi.VectorMath.createFrame(frameOrigin, dirU, dirV);
//            var outframe = new pointArray(false);
            outframe.pushFrame(startFrame);
        }
            this.setOutputData(1, outframe);      
    }
    LiteGraph.registerNodeType("Points2/getCrvSFrame", getCurveStartFrame);
    
function splitInComponents()
{

	this.addInput("pts","pointarray");
	this.addOutput("x","numarray");
	this.addOutput("y","numarray");
	this.addOutput("z","numarray");
	this.addOutput("Rx","numarray");
	this.addOutput("Ry","numarray");
	this.addOutput("Rz","numarray");
	this.addOutput("Scale","numarray");
//        this.addOutput("length","numarray");
        this.addOutput("xlength","numarray");
        this.addOutput("ylength","numarray");
        this.addOutput("zlength","numarray");

    this.properties = {};
}

splitInComponents.title = "splitPts";
splitInComponents.desc = "split Points in Components";

splitInComponents.prototype.onExecute = function()
{       
	var x=[],y=[],z=[],rx=[],ry=[],rz=[],scale=[];
//        var inp =  new pointArray();
        var inp = this.getInputData(0, new pointArray());
        for(var i=0;i<inp.getLength();i++)
	{var pt=inp.getElement(i);		
	x.push(pt.data[0]);
	y.push(pt.data[1]);
	z.push(pt.data[2]);
	rx.push(pt.data[3]);
	ry.push(pt.data[4]);
	rz.push(pt.data[5]);
	scale.push(pt.data[6]);
	}
	this.setOutputData(0, x);
	this.setOutputData(1, y);
	this.setOutputData(2, z);
	this.setOutputData(3, rx);
	this.setOutputData(4, ry);
	this.setOutputData(5, rz);
	this.setOutputData(6, scale);
//	this.setOutputData(7, [inp.length]);
	this.setOutputData(7, [inp.xlength]);
	this.setOutputData(8, [inp.ylength]);        
	this.setOutputData(9, [inp.zlength]);
}

LiteGraph.registerNodeType("Points2/splitInComponents", splitInComponents);
//

function PointExt()
{
//	this.minsize = [100, 0];
	this.addInput("X","numarray");
	this.addInput("Y","numarray");
	this.addInput("Z","numarray");
	this.addInput("AngleX","numarray");
	this.addInput("AngleY","numarray");
	this.addInput("AngleZ","numarray");
	this.addInput("pSize","numarray");
	this.addInput("xlength","numarray");
	this.addInput("ylength","numarray");
	this.addInput("zlength","numarray");
	
	this.addOutput("Point","pointarray");
	this.addOutput("Object","objectlist");
	
	this.properties = {mode:["Long","Long","Short"], x:[0], y:[0], z:[0], AngleX:[0], AngleY:[0], AngleZ:[0], pSize:[1]};
}

PointExt.title = "PointExt";
PointExt.desc = "PointExt";

PointExt.prototype.onExecute = function()
{
	var data = this.processInOut(this.properties.mode[0],
			this.multiProcess,
			this.properties.x,
			this.properties.y,
			this.properties.z,
			this.properties.AngleX,
			this.properties.AngleY,
			this.properties.AngleZ,
			this.properties.pSize);
			
	this.properties.x = data.inputs[0];
	this.properties.y = data.inputs[1];
	this.properties.z = data.inputs[2];
	this.properties.AngleX = data.inputs[3];
	this.properties.AngleY = data.inputs[4];
	this.properties.AngleZ = data.inputs[5];
	this.properties.pSize = data.inputs[6];
	if(this.inputs[7].link !== null&&this.inputs[8].link !== null&&this.inputs[9].link !== null){
            data.outputs[0].xlength= this.getInputData(7, "numarray")[0];
            data.outputs[0].ylength= this.getInputData(8, "numarray")[0];
            data.outputs[0].zlength= this.getInputData(9, "numarray")[0];
        }
	this.setOutputData(0, data.outputs[0]);
	this.setOutputData(1, data.outputs[1]);
}

PointExt.prototype.multiProcess = function(x,y,z,rx,ry,rz,scale)
{
	var out = [];
	//out[0] = new pointArray(true, x, y, z);
	
	out[0] = new pointArray();
	out[0].push(x,y,z,rx,ry,rz,scale);
					
	out[1] = factory('point', out[0].getPoint(0));
	return out;
}

LiteGraph.registerNodeType("Points2/PointExt", PointExt);

// bundlePts
    function bundlePts()
    {
        this.labels = "abcdef";
//        this.size = [60, 20];
//        this.dynamicInternal = ["a", "b", "c", "d", "e", "f"];
        this.internal = {a: [], b: []};
//        this.addOutput("Out", "numarray");
//this.properties={xl:0,yl:0,zl:0}
this.properties={}
//        this.zero=[0];
    }

    bundlePts.title = "bundlePts";
    bundlePts.desc = "bundlePts";

    bundlePts.prototype.onGetInputs = function ()
    {var inpName="";
        var list = [];
//        if(this.inputs[0].type==="numarray") inpName="Num";
        if(this.inputs[0].type==="pointarray") inpName="Pts";
//        if(this.inputs[0].type==="objectlist") inpName="Obj";

        list.push([inpName, this.inputs[0].type, {locked: true}]);

        return list;
    }

    bundlePts.prototype.onInputAdded = function (inp)
    {
        this.internal[inp.name] = 1;
        for (var i in this.inputs)
            this.inputs[i].locked = true;
        if (i > 1)
            this.inputs[i].locked = false;
    }

    bundlePts.prototype.onInputRemoved = function (slot, name)
    {
        if (this.internal[name])
            delete this.internal[name];
        for (var i in this.inputs)
            this.inputs[i].locked = true;
        if (this.inputs.length > 2)
            this.inputs[i].locked = false;
    }

    bundlePts.prototype.onAdded = function ()
    {
        for (var i = 0; i < this.labels.length; i++)
            if (this.internal[this.labels.charAt(i)] && this.inputs.length === i)
//            if (this.inputs.length === i)
                this.addInput("Pts", "pointarray");
        
            if(this.outputs.length===0) this.addOutput("Out", "pointarray");
        this.setDirtyCanvas(true);
    }

    bundlePts.prototype.onExecute = function ()
    {

            var inPts=[];
            var out = new pointArray();
            var minL=10000000;
            for (var i = 0; i < this.inputs.length; i++)
            {
                inPts.push(this.getInputData(i, new pointArray()));
                if(inPts[i].getLength()<minL)minL=inPts[i].getLength();
            }
//            for (var j = 0; j < inPts[0].getLength(); j++){
            for (var j = 0; j < minL; j++){
                for (var k = 0; k < inPts.length; k++){
                out.concat(inPts[k].getElement(j));
            }
            out.xlength=inPts.length;    
            out.ylength=1;            
//            out.zlength=inPts[0].getLength();
            out.zlength=minL;
            }            

        this.setOutputData(0, out);
    }

    LiteGraph.registerNodeType("Points2/bundlePts", bundlePts);
    
    // RotatePts
function RotatePts2()
{
	this.properties = { mode:["Long","Long","Short"], rX:[0], rY:[0], rZ:[0]};
	this.addInput("In", "pointarray");
	this.addInput("rX", "numarray");
	this.addInput("rY", "numarray");
	this.addInput("rZ", "numarray");
	this.addOutput("Out", "pointarray");
}

RotatePts2.title = "RotatePts2";
RotatePts2.desc = "RotatePts2";

RotatePts2.prototype.onExecute = function()
{
	var data = this.processInOut(this.properties.mode[0], this.multiProcess, new pointArray(true), this.properties.rX, this.properties.rY, this.properties.rZ);
	this.properties.rX = data.inputs[1];
	this.properties.rY = data.inputs[2];
	this.properties.rZ = data.inputs[3];
	this.setOutputData(0, data.outputs[0]);
}

RotatePts2.prototype.multiProcess = function(p, rX, rY, rZ)
{
	p.rotate(0,rX,rY,rZ);
	return [p];
}

LiteGraph.registerNodeType("Points2/RotatePts2", RotatePts2);

//
//////////////////////////
// node sPoint
// node sPoint based on the original Point node from Max Smirnov -
// the only diffrence is that a preselected point will be evaluated and the values
// will be preseted onAdded - the original point will be untouched
//
//////////////
function sPoint()
{
//	this.minsize = [100, 0];
	this.addInput("X","numarray");
	this.addInput("Y","numarray");
	this.addInput("Z","numarray");
	this.addOutput("Point","pointarray");
	this.addOutput("Object","objectlist");
	this.properties = {mode:["Long","Long","Short"], x:[0], y:[0], z:[0]};
}

sPoint.title = "Point";
sPoint.desc = "Point";

sPoint.prototype.onExecute = function()
{
	var data = this.processInOut(this.properties.mode[0], this.multiProcess, this.properties.x, this.properties.y, this.properties.z );
	this.properties.x = data.inputs[0];
	this.properties.y = data.inputs[1];
	this.properties.z = data.inputs[2];
	this.setOutputData(0, data.outputs[0]);
	this.setOutputData(1, data.outputs[1]);
}

sPoint.prototype.multiProcess = function(x,y,z)
{
	var out = [];
	out[0] = new pointArray(true, x, y, z);
	out[1] = factory('point', out[0].getPoint(0));
	return out;
}

sPoint.prototype.onAdded = function ()
    {	
        var inputPoint = moi.geometryDatabase.getSelectedObjects().getPoints();

        if (inputPoint.length > 0)
        {   this.properties.x[0]= inputPoint.item(0).pt.x;
            this.properties.y[0]= inputPoint.item(0).pt.y;
            this.properties.z[0]= inputPoint.item(0).pt.z;
            for (var i=1;i<inputPoint.length;i++){
            this.properties.x.push(inputPoint.item(i).pt.x);
            this.properties.y.push(inputPoint.item(i).pt.y);
            this.properties.z.push(inputPoint.item(i).pt.z);
        }
            moi.geometryDatabase.deselectAll();

        }


    }
LiteGraph.registerNodeType("Points2/sPoint", sPoint);
//

})();